using System;
using System.Drawing;
using System.Collections.Generic;
using System.Windows.Forms;
using Microsoft.DirectX;
using Direct3D = Microsoft.DirectX.Direct3D;

namespace DXCSharp
{
  // ---------------------------------------------------------------------
  // ---------------------------------------------------------------------
  public class Form1 : Form
  {
    #region Constants

    const int BoardW = 10;
    const int BoardH = 8;
    const int CellMax = 4;

    #endregion

    // -------------
    #region Members

    Direct3D.Device m_Device = null;
    Random m_Rand = new Random();
    int[,] m_Board;
    int m_nPieces;

    bool m_bPlayerTurn;
    bool m_bPlayerClicked;
    Point m_playerClick;

    
    #endregion

    // ---------------------------------------------------------------------
    // Form management
    // ---------------------------------------------------------------------
    #region Form management

    /// <summary>
    /// Constructor
    /// </summary>
    public Form1()
    {
      this.ClientSize = new System.Drawing.Size(400, 300);
      this.Text = "DXCSharp Test";

      ResetBoard();
    }
    #endregion

    // ---------------------------------------------------------------------
    // Graphics
    // ---------------------------------------------------------------------
    #region Graphics code

    static readonly int[,] BoardColors = 
      {
        { 0x000000, 0x303030 },
        { 0x484848, 0x181818 },
      };
    static readonly int[] PlayerColors =
      {
        0xFF5000,
        0x0050FF
      };

    /// <summary>
    /// Find the scale factor for a cell in the board
    /// </summary>
    public float BoardToClient()
    {
      float scale;
      if (ClientSize.Width / BoardW > ClientSize.Height / BoardH)
        scale = ClientSize.Height / BoardH;
      else
        scale = ClientSize.Width / BoardW;
      return scale;
    }

    /// <summary>
    /// Find the top-left corner of a cell in the board
    /// </summary>
    public Vector4 BoardToClient(int x, int y)
    {
      float scale, offsetX = 0, offsetY = 0;
      if (ClientSize.Width / BoardW > ClientSize.Height / BoardH)
      {
        scale = ClientSize.Height / BoardH;
        offsetX = (ClientSize.Width - scale * BoardW) / 2;
      }
      else
      {
        scale = ClientSize.Width / BoardW;
        offsetY = (ClientSize.Height - scale * BoardH) / 2;
      }
      return new Vector4(offsetX + scale * x, offsetY + scale * y, 0f, 1f);
    }

    /// <summary>
    /// Find the cell given the client coordinates
    /// </summary>
    public Point ClientToBoard(int x, int y)
    {
      float scale, offsetX = 0, offsetY = 0;
      if (ClientSize.Width / BoardW > ClientSize.Height / BoardH)
      {
        scale = ClientSize.Height / BoardH;
        offsetX = (ClientSize.Width - scale * BoardW) / 2;
      }
      else
      {
        scale = ClientSize.Width / BoardW;
        offsetY = (ClientSize.Height - scale * BoardH) / 2;
      }
      return new Point((int)((x - offsetX) / scale), (int)((y - offsetY) / scale));
    }

    /// <summary>
    /// Graphics initialization
    /// </summary>
    public bool InitializeGraphics()
    {
      try
      {
        // Now  setup our D3D stuff
        Direct3D.PresentParameters presentParams = new Direct3D.PresentParameters();
        presentParams.Windowed = true;
        presentParams.SwapEffect = Direct3D.SwapEffect.Discard;
        m_Device = new Direct3D.Device(0, Direct3D.DeviceType.Hardware, this,
            Direct3D.CreateFlags.SoftwareVertexProcessing, presentParams);
        return true;
      }
      catch (DirectXException)
      {
        return false;
      }
    }

    /// <summary>
    /// Rendering code
    /// </summary>
    public void Render()
    {
      if (m_Device == null)
        return;

      //Clear the backbuffer to a blue color 
      m_Device.Clear(Direct3D.ClearFlags.Target, System.Drawing.Color.FromArgb(255, m_Rand.Next(25), m_Rand.Next(25), m_Rand.Next(56)), 1.0f, 0);
      //Begin the scene
      m_Device.BeginScene();

      // Rendering of scene objects can happen here
      Direct3D.CustomVertex.TransformedColored[] av = new Direct3D.CustomVertex.TransformedColored[4];
      m_Device.VertexFormat = Direct3D.CustomVertex.TransformedColored.Format;

      for (int i = 0; i < BoardH; ++i)
        for (int j = 0; j < BoardW; ++j)
        {
          int k = (i + j) & 1;
          av[0].Position = BoardToClient(j, i);
          av[1].Position = BoardToClient(j+1, i);
          av[2].Position = BoardToClient(j+1, i+1);
          av[3].Position = BoardToClient(j, i+1);
          av[0].Color = BoardColors[k,0];
          av[1].Color = BoardColors[k,0];
          av[2].Color = BoardColors[k,1];
          av[3].Color = BoardColors[k,1];
          m_Device.DrawUserPrimitives(Direct3D.PrimitiveType.TriangleFan, 2, av);
        }
      for (int i = 0; i < BoardH; ++i)
        for (int j = 0; j < BoardW; ++j)
        {
          if (m_Board[i, j] != 0)
          {
            int c = PlayerColors[(m_Board[i, j] < 0) ? 0 : 1];
            float s = BoardToClient() / 2f;
            float r = s * Math.Abs(m_Board[i, j]) / CellMax;
            Vector4 vc = BoardToClient(j, i) + new Vector4(s, s, 0, 0);
            av[0].Position = vc - new Vector4(0, r, 0, 0);
            av[1].Position = vc + new Vector4(r, 0, 0, 0);
            av[2].Position = vc + new Vector4(0, r, 0, 0);
            av[3].Position = vc - new Vector4(r, 0, 0, 0);
            av[0].Color = c;
            av[1].Color = c;
            av[2].Color = c;
            av[3].Color = c;
            m_Device.DrawUserPrimitives(Direct3D.PrimitiveType.TriangleFan, 2, av);
          }
        }


      //End the scene
      m_Device.EndScene();
      m_Device.Present();
    }
    #endregion


    // ---------------------------------------------------------------------
    // Board management
    // ---------------------------------------------------------------------
    #region Board Management

    /// <summary>
    /// Reset the board to start a new game
    /// </summary>
    public void ResetBoard()
    {
      m_Board = new int[BoardH, BoardW];
      for (int i = 0; i < BoardH; ++i)
        for (int j = 0; j < BoardW; ++j)
          m_Board[i, j] = m_Rand.Next(CellMax*2-1) - (CellMax-1);
      m_nPieces = 0;

      m_bPlayerTurn = true;
      m_bPlayerClicked = false;
    }

    /// <summary>
    /// Checks if a cell exists in the board
    /// </summary>
    bool ValidCell(int x, int y)
    {
      return (x >= 0 && x < BoardW && y >= 0 && y < BoardH);
    }

    /// <summary>
    /// Process a cell invaded by the explosion of an adjacent cell
    /// </summary>
    void InvadeCell(int x, int y, int sign)
    {
      if (ValidCell(x, y))
        m_Board[y, x] = sign * (Math.Abs(m_Board[y, x]) + 1);
    }

    #endregion

    // ---------------------------------------------------------------------
    // Computer AI
    // ---------------------------------------------------------------------
    #region Computer AI

    /// <summary>
    /// Evaluate a cell for AI
    /// </summary>
    int ScoreAI(int x, int y, int cellAttack)
    {
      if (ValidCell(x, y))
      {
        if (m_Board[y, x] >= 0)
          if (cellAttack >= m_Board[y, x])
            return m_Board[y, x];
          else
            return -m_Board[y, x];
      }
      return 0;
    }

    /// <summary>
    /// Perform the computer AI
    /// </summary>
    void ComputerAI()
    {
      // The Computer AI goes here
      int x = 0, y = 0;
      int score = 0;

      // Find the cell that will cause the most damaging explosion
      for (int i = 0; i < BoardH; ++i)
        for (int j = 0; j < BoardW; ++j)
        {
          int v = Math.Max(0, -m_Board[i, j]);
          int s = v;
          s += ScoreAI(j + 1, i, v);
          s += ScoreAI(j - 1, i, v);
          s += ScoreAI(j, i + 1, v);
          s += ScoreAI(j, i - 1, v);
          if (s > score)
          {
            x = j;
            y = i;
            score = s;
          }
        }
      // If couldn't find a good cell to attack, choose at random
      if (score <= 0)
      {
        // Warning - infinite loop chance. Cheapo safeguard
        for (int i = 0; i < 10000; ++i)
        {
          x = m_Rand.Next(BoardW);
          y = m_Rand.Next(BoardH);
          if (m_Board[y, x] <= 0)
            break;
        }
      }
      if (m_Board[y, x] <= 0)
      {
        m_Board[y, x]--;
        m_nPieces++;
      }
    }

    #endregion

    // ---------------------------------------------------------------------
    // Game management
    // ---------------------------------------------------------------------
    #region Game Management

    /// <summary>
    /// Run a logic tick
    /// </summary>
    public void Run()
    {
      // Check if a cell must be blown up
      List<Point> LstExplodingCells = new List<Point>();
      for (int i = 0; i < BoardH; ++i)
        for (int j = 0; j < BoardW; ++j)
          if (Math.Abs(m_Board[i, j]) >= CellMax)
            LstExplodingCells.Add(new Point(j, i));
      if (LstExplodingCells.Count > 0)
      {
        // If so, select one at random
        Point p = LstExplodingCells[m_Rand.Next(LstExplodingCells.Count)];
        int sign = (m_Board[p.Y, p.X] < 0) ? -1 : 1;
        m_Board[p.Y, p.X] = 0;
        InvadeCell(p.X, p.Y - 1, sign);
        InvadeCell(p.X, p.Y + 1, sign);
        InvadeCell(p.X - 1, p.Y, sign);
        InvadeCell(p.X + 1, p.Y, sign);
      }
      else
      {
        // If nothing to explode and game hasn't ended, let players play
        if (m_bPlayerTurn)
        {
          // Let the player do his thing
          if (m_bPlayerClicked)
          {
            Point p = ClientToBoard(m_playerClick.X, m_playerClick.Y);
            m_bPlayerClicked = false;
            if (ValidCell(p.X, p.Y) && m_Board[p.Y, p.X] >= 0)
            {
              // Valid move
              m_Board[p.Y, p.X]++;
              m_nPieces++;
              m_bPlayerTurn = false;
            }
          }
        }
        else
        {
          ComputerAI();
          m_bPlayerTurn = true;
        }
      }
    }

    #endregion

    // ---------------------------------------------------------------------
    // Event handling
    // ---------------------------------------------------------------------
    #region Event Handling

    /// <summary>
    /// Handler for Painting
    /// </summary>
    protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
    {
      this.Render(); // Render on painting
    }

    /// <summary>
    /// Handler for key presses
    /// </summary>
    protected override void OnKeyPress(System.Windows.Forms.KeyPressEventArgs e)
    {
      if ((int)(byte)e.KeyChar == (int)System.Windows.Forms.Keys.Escape)
      {
        if (m_nPieces == 0)
          Close();
        else
          ResetBoard();
      }
    }

    /// <summary>
    /// Handler for mouse clicks
    /// </summary>
    protected override void OnMouseDown(System.Windows.Forms.MouseEventArgs e)
    {
      // Find where did the user click
      if (m_bPlayerTurn && e.Button == MouseButtons.Left)
      {
        m_playerClick = e.Location;
        m_bPlayerClicked = true;
      }
    }
    #endregion
  }

  // ---------------------------------------------------------------------
  // ---------------------------------------------------------------------

  // ---------------------------------------------------------------------
  // Main Loop
  // ---------------------------------------------------------------------
  #region Main Loop

  class Program
  {
    /// <summary>
    /// The main entry point for the application.
    /// </summary>
    static void Main()
    {

      using (Form1 frm = new Form1())
      {
        if (!frm.InitializeGraphics()) // Initialize Direct3D
        {
          MessageBox.Show("Could not initialize Direct3D.  This tutorial will exit.");
          return;
        }
        frm.Show();

        // While the form is still valid, render and process messages
        while (frm.Created)
        {
          frm.Run();
          frm.Render();
          Application.DoEvents();
          System.Threading.Thread.Sleep(50);
        }
      }
    }
  }
  #endregion
}